@@ -1,4 +1,4 @@
-// Copyright (c) 2017-2023 Advanced Micro Devices, Inc. All rights reserved.
+// Copyright (c) 2017-2024 Advanced Micro Devices, Inc. All rights reserved.
 //
 // Permission is hereby granted, free of charge, to any person obtaining a copy
 // of this software and associated documentation files (the "Software"), to deal
@@ -30,7 +30,6 @@
 #include "detail/lookback_scan_state.hpp"
 
 #include "../config.hpp"
-#include "../detail/match_result_type.hpp"
 #include "../detail/temp_storage.hpp"
 #include "../detail/various.hpp"
 #include "../functional.hpp"
@@ -215,14 +214,20 @@ hipError_t reduce_by_key_impl(void*                     temporary_storage,
     // The running accumulation across the launch boundary.
     accumulator_type* d_previous_accumulated = nullptr;
 
+    detail::temp_storage::layout layout{};
+    const hipError_t             layout_result
+        = scan_state_type::get_temp_storage_layout(number_of_tiles, stream, layout);
+    if(layout_result != hipSuccess)
+    {
+        return layout_result;
+    }
+
     const hipError_t partition_result = detail::temp_storage::partition(
         temporary_storage,
         storage_size,
         detail::temp_storage::make_linear_partition(
             // This is valid even with scan_state_with_sleep_type
-            detail::temp_storage::make_partition(
-                &scan_state_storage,
-                scan_state_type::get_temp_storage_layout(number_of_tiles)),
+            detail::temp_storage::make_partition(&scan_state_storage, layout),
             detail::temp_storage::make_partition(&ordered_bid_storage,
                                                  ordered_tile_id_type::get_temp_storage_layout()),
             detail::temp_storage::ptr_aligned_array(&d_global_head_count, use_limited_size ? 1 : 0),
@@ -239,12 +244,23 @@ hipError_t reduce_by_key_impl(void*                     temporary_storage,
     {
         return result;
     }
+
+    scan_state_type scan_state{};
+    hipError_t      scan_state_result
+        = scan_state_type::create(scan_state, scan_state_storage, number_of_tiles, stream);
+    scan_state_with_sleep_type scan_state_with_sleep{};
+    scan_state_result = scan_state_with_sleep_type::create(scan_state_with_sleep,
+                                                           scan_state_storage,
+                                                           number_of_tiles,
+                                                           stream);
+
+    if(scan_state_result != hipSuccess)
+    {
+        return scan_state_result;
+    }
+
     auto with_scan_state
-        = [use_sleep,
-           scan_state = scan_state_type::create(scan_state_storage, number_of_tiles),
-           scan_state_with_sleep
-           = scan_state_with_sleep_type::create(scan_state_storage, number_of_tiles)](
-              auto&& func) mutable -> decltype(auto)
+        = [use_sleep, scan_state, scan_state_with_sleep](auto&& func) mutable -> decltype(auto)
     {
         if(use_sleep)
         {
@@ -371,8 +387,11 @@ hipError_t reduce_by_key_impl(void*                     temporary_storage,
 ///
 /// \par Overview
 /// * Supports non-commutative reduction operators. However, a reduction operator should be
-/// associative. When used with non-associative functions the results may be non-deterministic
-/// and/or vary in precision.
+/// associative.
+/// * When used with non-associative functions (e.g. floating point arithmetic operations):
+///    - the results may be non-deterministic and/or vary in precision,
+///    - and bit-wise reproducibility is not guaranteed, that is, results from multiple runs
+///      using the same input values on the same device may not be bit-wise identical.
 /// * Returns the required size of \p temporary_storage in \p storage_size
 /// if \p temporary_storage in a null pointer.
 /// * Ranges specified by \p keys_input and \p values_input must have at least \p size elements.
@@ -380,8 +399,7 @@ hipError_t reduce_by_key_impl(void*                     temporary_storage,
 /// * Ranges specified by \p unique_output and \p aggregates_output must have at least
 /// <tt>*unique_count_output</tt> (i.e. the number of unique keys) elements.
 ///
-/// \tparam Config - [optional] configuration of the primitive. It can be `reduce_by_key_config_v2`
-/// or `default_config`
+/// \tparam Config - [optional] configuration of the primitive. It has to be `reduce_by_key_config` or a class derived from it.
 /// \tparam KeysInputIterator - random-access iterator type of the input range. Must meet the
 /// requirements of a C++ InputIterator concept. It can be a simple pointer type.
 /// \tparam ValuesInputIterator - random-access iterator type of the input range. Must meet the
